from test.testbase import DbIsolatedTest, run_as_sync

from databases.interfaces import Record

from db.python.layers.participant import ParticipantLayer
from models.models import PRIMARY_EXTERNAL_ORG, ParticipantUpsertInternal


class TestImportIndividualMetadata(DbIsolatedTest):
    """Test importing individual metadata"""

    @run_as_sync
    async def test_import_many_hpo_terms(self):
        """Test import hpo terms from many columns"""
        pl = ParticipantLayer(self.connection)

        await pl.upsert_participant(
            ParticipantUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'TP01'})
        )

        headers = [
            'Individual ID',
            'HPO Term 1',
            'HPO Term 2',
            'HPO Term 3',
            'HPO Term 20',
        ]
        rows_to_insert = [
            ['TP01', 'HP:0000001', 'HP:0000002', 'HP:0000003', 'HP:0000004']
        ]

        await pl.generic_individual_metadata_importer(headers, rows_to_insert)

        db_rows: list[Record] = list(
            await self.connection.connection.fetch_all(
                'SELECT participant_id, description, value FROM participant_phenotypes'
            )
        )

        self.assertEqual(1, len(db_rows))
        self.assertEqual('HPO Terms (present)', db_rows[0]['description'])
        self.assertEqual(
            '"HP:0000001,HP:0000002,HP:0000003,HP:0000004"', db_rows[0]['value']
        )

    @run_as_sync
    async def test_import_basic_metadata(self):
        """Test basic data for 2 participants and 2 columns"""
        pl = ParticipantLayer(self.connection)

        await pl.upsert_participants(
            [
                ParticipantUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'TP01'}),
                ParticipantUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'TP02'}),
            ]
        )

        headers = ['Individual ID', 'HPO Term 20', 'Age of Onset']
        rows_to_insert = [
            ['TP01', 'HP:0000020', 'Congenital'],
            ['TP02', 'HP:00000021; HP:023', 'Infantile'],
        ]

        await pl.generic_individual_metadata_importer(headers, rows_to_insert)

        rows = list(
            await self.connection.connection.fetch_all(
                'SELECT participant_id, description, value FROM participant_phenotypes'
            )
        )

        self.assertEqual(4, len(rows))

        first_p_rows = rows[:2]
        second_p_rows = rows[2:]

        self.assertEqual('Age of Onset', first_p_rows[0]['description'])
        self.assertEqual('"Congenital onset"', first_p_rows[0]['value'])
        self.assertEqual('HPO Terms (present)', first_p_rows[1]['description'])
        self.assertEqual('"HP:0000020"', first_p_rows[1]['value'])

        self.assertEqual('Age of Onset', second_p_rows[0]['description'])
        self.assertEqual('"Infantile onset"', second_p_rows[0]['value'])
        self.assertEqual('HPO Terms (present)', second_p_rows[1]['description'])
        self.assertEqual('"HP:00000021,HP:023"', second_p_rows[1]['value'])



from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

from db.python.layers.sample import SampleLayer
from metamist.parser.generic_parser import (
    PRIMARY_EXTERNAL_ORG,
    GenericParser,
    ParsedSample,
)
from models.models.sample import SampleUpsertInternal
from models.utils.sample_id_format import sample_id_transform_to_raw


class GenericParserForTest(GenericParser):
    """
    There's an abstract method on the GenericParser, so just implement a dummy
    """

    async def get_assays_from_group(self, *_, **__) -> list:
        """Dummy method to implement abstract method"""
        return []

    def get_primary_sample_id(self, *_, **__) -> str:
        """Dummy method to implement abstract method"""
        return ''

    def get_primary_participant_id(self, *_, **__) -> str | None:
        """Dummy method to implement abstract method"""
        return ''

    def has_participants(self, *_, **__) -> bool:
        """Dummy method to implement abstract method"""
        return True


class TestGenericParser(DbIsolatedTest):
    """Test generic parser specific methods"""

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_nested_samples_misc(self, mock_graphql_query):
        """
        Test a bunch of things related to nested samples
        They're all sort of sequential, so just do them all in one test
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        slayer = SampleLayer(self.connection)
        u_sample = await slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'},
                type='blood',
                meta={},
                nested_samples=[
                    # intentionally don't have EX01.1
                    # SampleUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'EX01.1'),
                    SampleUpsertInternal(
                        external_ids={PRIMARY_EXTERNAL_ORG: 'EX01.2'},
                        type='blood',
                        meta={},
                    )
                ],
            )
        )

        u_child2 = u_sample.nested_samples[0]

        p = GenericParserForTest(
            path_prefix=None, search_paths=[], project=self.project_name
        )

        def _make_parsed_sample(exid):
            return ParsedSample(
                participant=None,
                rows=[],
                internal_sid=None,
                external_sids={PRIMARY_EXTERNAL_ORG: exid},
                sample_type='blood',
                meta={},
            )

        nested_sample = _make_parsed_sample('EX01')
        child_1 = _make_parsed_sample('EX01.1')
        child_2 = _make_parsed_sample('EX01.2')
        nested_sample.samples.extend([child_1, child_2])

        # Test 1: test the matching works for all samples including nested
        await p.match_sample_ids([nested_sample])

        self.assertEqual(
            u_sample.id, sample_id_transform_to_raw(nested_sample.internal_sid)
        )
        self.assertIsNone(child_1.internal_sid)
        self.assertEqual(u_child2.id, sample_id_transform_to_raw(child_2.internal_sid))

        # Test 2: test the summary generated is correctly counting for nested samples
        summary = p.prepare_summary([], [nested_sample], [], [])
        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(2, summary.samples.update)

        # Test 3: test the to_sm method works for nested samples
        transport_model = nested_sample.to_sm()
        self.assertEqual(transport_model.id, nested_sample.internal_sid)
        self.assertEqual(transport_model.nested_samples[0].id, child_1.internal_sid)
        self.assertEqual(transport_model.nested_samples[1].id, child_2.internal_sid)



# pylint: disable=protected-access
import datetime
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from typing import Any
from unittest import mock

import google.cloud.bigquery as bq

from db.python.tables.bq.billing_daily import BillingDailyTable
from db.python.tables.bq.billing_filter import BillingFilter
from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from db.python.utils import InternalError
from models.models import BillingColumn, BillingTotalCostQueryModel


class TestBillingDailyTable(BqTest):
    """Test BillingRawTable and its methods"""

    def setUp(self):
        super().setUp()

        # setup table object
        self.table_obj = BillingDailyTable(self.connection)

    def test_query_to_partitioned_filter(self):
        """Test query to partitioned filter conversion"""

        # given
        start_date = '2023-01-01'
        end_date = '2024-01-01'
        filters: dict[BillingColumn, str | list[Any] | dict[Any, Any]] = {
            BillingColumn.TOPIC: 'TEST_TOPIC'
        }

        # expected
        expected_filter = BillingFilter(
            day=GenericBQFilter(
                gte=datetime.datetime(2023, 1, 1, 0, 0),
                lte=datetime.datetime(2024, 1, 1, 0, 0),
            ),
            topic=GenericBQFilter(eq='TEST_TOPIC'),
        )

        query = BillingTotalCostQueryModel(
            fields=[],  # not relevant for this test, but can't be null generally
            start_date=start_date,
            end_date=end_date,
            filters=filters,
        )
        filter_ = BillingDailyTable._query_to_partitioned_filter(query)

        # BillingFilter has __eq__ method, so we can compare them directly
        self.assertEqual(expected_filter, filter_)

    def test_error_no_connection(self):
        """Test No connection exception"""

        with self.assertRaises(InternalError) as context:
            BillingDailyTable(None)

        self.assertTrue(
            "No connection was provided to the table 'BillingDailyTable'"
            in str(context.exception)
        )

    def test_get_table_name(self):
        """Test get_table_name"""

        # table name is set in the class
        given_table_name = 'TEST_TABLE_NAME'

        # set table name
        self.table_obj.table_name = given_table_name

        # test get table name function
        table_name = self.table_obj.get_table_name()

        self.assertEqual(given_table_name, table_name)

    @run_as_sync
    async def test_last_loaded_day_return_valid_day(self):
        """Test _last_loaded_day"""

        given_last_day = '2021-01-01 00:00:00'

        # mock BigQuery result

        self.bq_result.result.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                last_loaded_day=datetime.datetime.strptime(
                    given_last_day, '%Y-%m-%d %H:%M:%S'
                ),
            )
        ]  # 2021-01-01

        # test get table name function
        last_loaded_day = await self.table_obj._last_loaded_day()

        self.assertEqual(given_last_day, last_loaded_day)

    @run_as_sync
    async def test_last_loaded_day_return_none(self):
        """Test _last_loaded_day as None"""

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get table name function
        last_loaded_day = await self.table_obj._last_loaded_day()

        self.assertEqual(None, last_loaded_day)

    def test_prepare_daily_cost_subquery(self):
        """Test _prepare_daily_cost_subquery"""

        self.table_obj.table_name = 'TEST_TABLE_NAME'

        # given
        given_field = BillingColumn.COST
        given_query_params: list[Any] = []
        given_last_loaded_day = '2021-01-01 00:00:00'

        (
            query_params,
            daily_cost_field,
            daily_cost_join,
        ) = self.table_obj._prepare_daily_cost_subquery(
            given_field, given_query_params, given_last_loaded_day
        )

        # expected
        expected_daily_cost_join = """LEFT JOIN (
            SELECT
                cost as field,
                cost_category,
                SUM(cost) as cost
            FROM
            `TEST_TABLE_NAME`
            WHERE day = TIMESTAMP(@last_loaded_day)
            GROUP BY
                field,
                cost_category
        ) day
        ON month.field = day.field
        AND month.cost_category = day.cost_category
        """

        self.assertEqual(
            [
                bq.ScalarQueryParameter(
                    'last_loaded_day', 'STRING', '2021-01-01 00:00:00'
                )
            ],
            query_params,
        )

        self.assertEqual(', day.cost as daily_cost', daily_cost_field)
        self.assertEqual(expected_daily_cost_join, daily_cost_join)

    @run_as_sync
    async def test_get_entities_as_empty_list(self):
        """
        Test get_topics, get_invoice_months,
        get_cost_categories and get_skus as empty list
        """

        # mock BigQuery result as empty list
        self.bq_result.result.return_value = []

        # test get_topics function
        records = await self.table_obj.get_topics()
        self.assertEqual([], records)

        # test get_invoice_months function
        records = await self.table_obj.get_invoice_months()
        self.assertEqual([], records)

        # test get_cost_categories function
        records = await self.table_obj.get_cost_categories()
        self.assertEqual([], records)

        # test get_skus function
        records = await self.table_obj.get_skus()
        self.assertEqual([], records)

    @run_as_sync
    async def test_get_topics_return_valid_list(self):
        """Test get_topics as empty list"""

        # mock BigQuery result as list of 2 records
        self.bq_result.result.return_value = [
            {'topic': 'TOPIC1'},
            {'topic': 'TOPIC2'},
        ]

        # test get_topics function
        records = await self.table_obj.get_topics()

        self.assertEqual(['TOPIC1', 'TOPIC2'], records)

    @run_as_sync
    async def test_get_invoice_months_return_valid_list(self):
        """Test get_invoice_months as empty list"""

        # mock BigQuery result as list of 2 records
        self.bq_result.result.return_value = [
            {'invoice_month': '202401'},
            {'invoice_month': '202402'},
        ]

        # test get_invoice_months function
        records = await self.table_obj.get_invoice_months()

        self.assertEqual(['202401', '202402'], records)

    @run_as_sync
    async def test_get_cost_categories_return_valid_list(self):
        """Test get_cost_categories as empty list"""

        # mock BigQuery result as list of 2 records
        self.bq_result.result.return_value = [
            {'cost_category': 'CAT1'},
        ]

        # test get_cost_categories function
        records = await self.table_obj.get_cost_categories()

        self.assertEqual(['CAT1'], records)

    @run_as_sync
    async def test_get_skus_return_valid_list(self):
        """Test get_skus as empty list"""

        # mock BigQuery result as list of 3 records
        self.bq_result.result.return_value = [
            {'sku': 'SKU1'},
            {'sku': 'SKU2'},
            {'sku': 'SKU3'},
        ]

        # test get_skus function
        records = await self.table_obj.get_skus()
        self.assertEqual(['SKU1', 'SKU2', 'SKU3'], records)

        # test get_skus function with limit,
        # limit is ignored in the test as we already have mockup data
        records = await self.table_obj.get_skus(limit=3)
        self.assertEqual(['SKU1', 'SKU2', 'SKU3'], records)

        # test get_skus function with limit & offset
        # limit & offset are ignored in the test as we already have mockup data
        records = await self.table_obj.get_skus(limit=3, offset=1)
        self.assertEqual(['SKU1', 'SKU2', 'SKU3'], records)



# pylint: disable=protected-access too-many-public-methods
import datetime
import json
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from unittest.mock import patch

from api.routes import billing
from models.models import (
    AnalysisCostRecord,
    BillingColumn,
    BillingCostBudgetRecord,
    BillingTotalCostQueryModel,
    BillingTotalCostRecord,
)

TEST_API_BILLING_USER = 'test_user'


class TestApiBilling(BqTest):
    """
    Test API Billing routes
    Billing routes are only calling layer functions and returning data it received
    This set of tests only checks if all routes code has been called
    and if data returned is the same as data received from layer functions
    It does not check all possible combination of parameters as
    Billing Layer and BQ Tables should be testing those
    """

    def setUp(self):
        super().setUp()

        # make billing enabled by default for all the calls
        patcher = patch('api.routes.billing.is_billing_enabled', return_value=True)
        self.mockup_is_billing_enabled = patcher.start()

    @run_as_sync
    @patch('api.routes.billing.is_billing_enabled', return_value=False)
    async def test_get_gcp_projects_no_billing(self, _mockup_is_billing_enabled):
        """
        Test get_gcp_projects function
        This function should raise ValueError if billing is not enabled
        """
        with self.assertRaises(ValueError) as context:
            await billing.get_gcp_projects('test_user')

        self.assertTrue('Billing is not enabled' in str(context.exception))

    @run_as_sync
    @patch('api.routes.billing._get_billing_layer_from')
    @patch('db.python.layers.billing.BillingLayer.get_cost_by_ar_guid')
    async def test_get_cost_by_ar_guid(
        self, mock_get_cost_by_ar_guid, mock_get_billing_layer
    ):
        """
        Test get_cost_by_ar_guid function
        """
        ar_guid = 'test_ar_guid'
        mockup_record_json = {
            'total': {
                'ar_guid': ar_guid,
                'cost': 0.0,
                'usage_end_time': None,
                'usage_start_time': None,
            },
            'topics': [],
            'categories': [],
            'batches': [],
            'skus': [],
            'seq_groups': [],
            'wdl_tasks': [],
            'cromwell_sub_workflows': [],
            'cromwell_workflows': [],
            'dataproc': [],
        }

        mockup_record = [AnalysisCostRecord.from_dict(mockup_record_json)]
        mock_get_billing_layer.return_value = self.layer
        mock_get_cost_by_ar_guid.return_value = mockup_record
        response = await billing.get_cost_by_ar_guid(
            ar_guid, author=TEST_API_BILLING_USER
        )
        resp_json = json.loads(response.body.decode('utf-8'))
        self.assertEqual(1, len(resp_json))

        self.assertDictEqual(mockup_record_json, resp_json[0])

    @run_as_sync
    @patch('api.routes.billing._get_billing_layer_from')
    @patch('db.python.layers.billing.BillingLayer.get_cost_by_batch_id')
    async def test_get_cost_by_batch_id(
        self, mock_get_cost_by_batch_id, mock_get_billing_layer
    ):
        """
        Test get_cost_by_batch_id function
        """
        ar_guid = 'test_ar_guid'
        batch_id = 'test_batch_id'
        mockup_record_json = {
            'total': {
                'ar_guid': ar_guid,
                'cost': 0.0,
                'usage_end_time': None,
                'usage_start_time': None,
            },
            'topics': [],
            'categories': [],
            'batches': [
                {
                    'batch_id': batch_id,
                    'batch_name': None,
                    'cost': 0.0,
                    'usage_start_time': datetime.datetime.now().isoformat(),
                    'usage_end_time': datetime.datetime.now().isoformat(),
                    'jobs_cnt': 0,
                    'skus': [],
                    'jobs': [],
                    'seq_groups': [],
                }
            ],
            'skus': [],
            'seq_groups': [],
            'wdl_tasks': [],
            'cromwell_sub_workflows': [],
            'cromwell_workflows': [],
            'dataproc': [],
        }

        mockup_record = [AnalysisCostRecord.from_dict(mockup_record_json)]
        mock_get_billing_layer.return_value = self.layer
        mock_get_cost_by_batch_id.return_value = mockup_record
        response = await billing.get_cost_by_batch_id(
            batch_id, author=TEST_API_BILLING_USER
        )
        resp_json = json.loads(response.body.decode('utf-8'))

        self.assertEqual(1, len(resp_json))
        self.assertDictEqual(mockup_record_json, resp_json[0])

    @run_as_sync
    @patch('api.routes.billing._get_billing_layer_from')
    @patch('db.python.layers.billing.BillingLayer.get_total_cost')
    async def test_get_total_cost(self, mock_get_total_cost, mock_get_billing_layer):
        """
        Test get_total_cost function
        """
        query = BillingTotalCostQueryModel(fields=[], start_date='', end_date='')
        mockup_record = [{'cost': 123.45}, {'cost': 123}]
        expected = [BillingTotalCostRecord.from_json(r) for r in mockup_record]
        mock_get_billing_layer.return_value = self.layer
        mock_get_total_cost.return_value = mockup_record
        records = await billing.get_total_cost(query, author=TEST_API_BILLING_USER)
        self.assertEqual(expected, records)

    @run_as_sync
    @patch('api.routes.billing._get_billing_layer_from')
    @patch('db.python.layers.billing.BillingLayer.get_running_cost')
    async def test_get_running_cost(
        self, mock_get_running_cost, mock_get_billing_layer
    ):
        """
        Test get_running_cost function
        """
        mockup_record = [
            BillingCostBudgetRecord.from_json(
                {'field': 'TOPIC1', 'total_monthly': 999.99, 'details': []}
            ),
            BillingCostBudgetRecord.from_json(
                {'field': 'TOPIC2', 'total_monthly': 123, 'details': []}
            ),
        ]
        mock_get_billing_layer.return_value = self.layer
        mock_get_running_cost.return_value = mockup_record
        records = await billing.get_running_costs(
            field=BillingColumn.TOPIC,
            invoice_month=None,
            source=None,
            author=TEST_API_BILLING_USER,
        )
        self.assertEqual(mockup_record, records)

    @patch('api.routes.billing._get_billing_layer_from')
    async def call_api_function(
        self,
        api_function,
        mock_layer_function,
        mock_get_billing_layer=None,
    ):
        """
        Common wrapper for all API calls, to avoid code duplication
        API function is called with author=TEST_API_BILLING_USER
        get_author function will be tested separately
        We only testing if routes are calling layer functions and
        returning data it received
        """
        mock_get_billing_layer.return_value = self.layer
        mockup_records = ['RECORD1', 'RECORD2']
        mock_layer_function.return_value = mockup_records
        records = await api_function(author=TEST_API_BILLING_USER)
        self.assertEqual(mockup_records, records)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_gcp_projects')
    async def test_get_gcp_projects(self, mock_get_gcp_projects):
        """
        Test get_gcp_projects function
        """
        await self.call_api_function(billing.get_gcp_projects, mock_get_gcp_projects)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_topics')
    async def test_get_topics(self, mock_get_topics):
        """
        Test get_topics function
        """
        await self.call_api_function(billing.get_topics, mock_get_topics)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_cost_categories')
    async def test_get_cost_categories(self, mock_get_cost_categories):
        """
        Test get_cost_categories function
        """
        await self.call_api_function(
            billing.get_cost_categories, mock_get_cost_categories
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_skus')
    async def test_get_skus(self, mock_get_skus):
        """
        Test get_skus function
        """
        await self.call_api_function(billing.get_skus, mock_get_skus)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_datasets')
    async def test_get_datasets(self, mock_get_datasets):
        """
        Test get_datasets function
        """
        await self.call_api_function(billing.get_datasets, mock_get_datasets)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_sequencing_types')
    async def test_get_sequencing_types(self, mock_get_sequencing_types):
        """
        Test get_sequencing_types function
        """
        await self.call_api_function(
            billing.get_sequencing_types, mock_get_sequencing_types
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_stages')
    async def test_get_stages(self, mock_get_stages):
        """
        Test get_stages function
        """
        await self.call_api_function(billing.get_stages, mock_get_stages)

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_sequencing_groups')
    async def test_get_sequencing_groups(self, mock_get_sequencing_groups):
        """
        Test get_sequencing_groups function
        """
        await self.call_api_function(
            billing.get_sequencing_groups, mock_get_sequencing_groups
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_compute_categories')
    async def test_get_compute_categories(self, mock_get_compute_categories):
        """
        Test get_compute_categories function
        """
        await self.call_api_function(
            billing.get_compute_categories, mock_get_compute_categories
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_cromwell_sub_workflow_names')
    async def test_get_cromwell_sub_workflow_names(
        self, mock_get_cromwell_sub_workflow_names
    ):
        """
        Test get_cromwell_sub_workflow_names function
        """
        await self.call_api_function(
            billing.get_cromwell_sub_workflow_names,
            mock_get_cromwell_sub_workflow_names,
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_wdl_task_names')
    async def test_get_wdl_task_names(self, mock_get_wdl_task_names):
        """
        Test get_wdl_task_names function
        """
        await self.call_api_function(
            billing.get_wdl_task_names, mock_get_wdl_task_names
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_invoice_months')
    async def test_get_invoice_months(self, mock_get_invoice_months):
        """
        Test get_invoice_months function
        """
        await self.call_api_function(
            billing.get_invoice_months, mock_get_invoice_months
        )

    @run_as_sync
    @patch('db.python.layers.billing.BillingLayer.get_namespaces')
    async def test_get_namespaces(self, mock_get_namespaces):
        """
        Test get_namespaces function
        """
        await self.call_api_function(billing.get_namespaces, mock_get_namespaces)



#!/usr/bin/env python

"""
This script will take a path to a csv file, containing a column with rows populated
by paths to delete. The --delete-field-name option allows you to specify which field
of the csv contains the paths to delete.
Creates a log file of all the deletes, in the same directory as the input csv.

The typical upload-bucket audit results will be found in a file such as:
 > gs://cpg-dataset-main-upload/audit_results/2023-xx-xx/dataset_seqtype_readstype_assay_files_to_delete_20xx-xx-xx.csv
And the field name containing the delete paths will be
 > "Assay_Read_File_Path"
"""

import csv
import logging
import os
import sys
from datetime import datetime

import click
from cloudpathlib import AnyPath, CloudPath
from google.cloud import storage

from cpg_utils.config import get_config

from metamist.audit.audithelper import AuditHelper

CLIENT = storage.Client()

TODAY = datetime.today().strftime('%Y-%m-%d')


def clean_up_cloud_storage(locations: list[CloudPath]):
    """Given a list of locations of files to be deleted"""
    deleted_files = []
    for location in locations:
        try:
            location.unlink()
            logging.info(f'{location.name} was deleted from cloud storage.')
            deleted_files.append(location.name)
        # Many possible http exceptions could occur so use a broad exception
        except Exception:  # pylint: disable=W0718
            logging.warning(f'{location.name} threw an exception - file not deleted.')

    return deleted_files


@click.command('Delete the assay files found by the audit')
@click.option(
    '--delete-field-name',
    '-d',
    help='The field in the input csv corresponding to the files to delete',
    required=True,
)
@click.argument('delete-file-path')
def main(delete_field_name, delete_file_path):
    """
    Inputs:
        - The name of the field containing the paths to delete in the input csv
        - The full GSPath to the csv containing the paths to delete
    """
    access_level = get_config()['workflow']['access_level']

    if access_level == 'standard':
        raise ValueError(
            'Standard cannot be used as the access level. Full or test is required.'
        )

    paths = set()
    with AnyPath(delete_file_path).open('r') as f:  # pylint: disable=E1101
        logging.info(f'Getting file {delete_file_path}...')
        reader = csv.DictReader(f)
        for row in reader:
            paths.add(CloudPath(row[delete_field_name]))

    deleted_files = clean_up_cloud_storage(list(paths))
    logging.info(f'{len(deleted_files)} sequence files deleted.')

    # Write a log of the deleted files to the same location
    log_path = f'{delete_file_path.removesuffix(os.path.basename(delete_file_path))}deleted_{TODAY}.csv'
    AuditHelper.write_csv_report_to_cloud(
        deleted_files, log_path, header_row=['Deleted_file_path']
    )


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(levelname)s %(module)s:%(lineno)d - %(message)s',
        datefmt='%Y-%M-%d %H:%M:%S',
        stream=sys.stderr,
    )

    main()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python

"""
Setup script for the Python package. Dependencies for server are listed separately
in requirements.txt, dev dependencies are listed in requirements-dev.txt.
"""

from setuptools import find_packages, setup

PKG = 'metamist'

all_packages = ['metamist']
all_packages.extend('metamist.' + p for p in sorted(find_packages('./metamist')))

with open('README.md', encoding='utf-8') as f:
    readme = f.read()


setup(
    name=PKG,
    # This tag is automatically updated by bump2version
    version='7.4.1',
    description='Python API for interacting with the Sample API system',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/populationgenomics/metamist',
    license='MIT',
    packages=all_packages,
    install_requires=[
        'backoff>=2.2.1',
        'click',
        'google-auth',
        'google-api-core',  # dependency to google-auth that however is not
        # pulled automatically: https://github.com/googleapis/google-auth-library-python/blob/main/setup.py#L22-L27
        'urllib3 >= 1.25.3',
        'python-dateutil',
        'requests',
        'typing-extensions',
        # for get id-token
        'cpg-utils >= 5.0.5',
        'gql[aiohttp,requests]',
        'tabulate >= 0.9.0',
    ],
    entry_points={
        'metamist_parser': [
            'GenericMetadataParser = metamist.parser.generic_metadata_parser:GenericMetadataParser',
            'SampleFileMapParser = metamist.parser.sample_file_map_parser:SampleFileMapParser',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    keywords='bioinformatics',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



from datetime import datetime
from io import StringIO
from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

from db.python.layers import ParticipantLayer
from metamist.parser.generic_parser import ParsedParticipant
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
)
from scripts.parse_existing_cohort import Columns, ExistingCohortParser


class TestExistingCohortParser(DbIsolatedTest):
    """Test the ExistingCohortParser"""

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.datetime_added')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_single_row(
        self,
        mock_filesize,
        mock_fileexists,
        mock_datetime_added,
        mock_graphql_query,
    ):
        """
        Test importing a single row, forms objects and checks response
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_filesize.return_value = 111
        mock_fileexists.return_value = False
        mock_datetime_added.return_value = datetime.fromisoformat('2022-02-02T22:22:22')

        rows = [
            'HEADER',
            '""',
            'Application\tExternal ID\tSample Concentration (ng/ul)\tVolume (uL)\tSex\tSample/Name\tReference Genome\tParticipant ID\t',
            'App\tEXTID1234\t100\t100\tFemale\t220405_FLUIDX1234\thg38\tPID123',
        ]
        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=False,
            sequencing_type='genome',
        )

        parser.filename_map = {
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
        }

        file_contents = '\n'.join(rows)
        participants: list[ParsedParticipant]
        summary, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(0, summary.assays.update)

        sample_to_add = participants[0].samples[0]
        self.assertEqual('EXTID1234', sample_to_add.primary_external_id)
        expected_sequence_dict = {
            'reference_genome': 'hg38',
            'platform': 'App',
            'concentration': 100,
            'volume': 100,
            'fluid_x_tube_id': '220405_FLUIDX1234',
            'reads_type': 'fastq',
            'reads': [
                {
                    'location': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
                    'basename': 'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
                    'class': 'File',
                    'checksum': None,
                    'size': 111,
                    'datetime_added': '2022-02-02T22:22:22',
                },
                {
                    'location': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
                    'basename': 'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
                    'class': 'File',
                    'checksum': None,
                    'size': 111,
                    'datetime_added': '2022-02-02T22:22:22',
                },
            ],
            'sequencing_platform': 'illumina',
            'sequencing_technology': 'short-read',
            'sequencing_type': 'genome',
            'batch': 'M01',
        }
        assay = sample_to_add.sequencing_groups[0].assays[0]
        self.maxDiff = None
        self.assertDictEqual(expected_sequence_dict, assay.meta)
        return

    @run_as_sync
    async def test_no_header(self):
        """
        Test input without a header
        """

        rows = [
            'Application\tExternal ID\tSample Concentration (ng/ul)\tVolume (uL)\tSex\tSample/Name\tReference Genome\tParticipant ID\t',
            'App\tEXTID1234\t100\t100\tFemale\t220405_FLUIDX1234\thg38\tPID123',
        ]
        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=False,
            sequencing_type='genome',
        )

        parser.filename_map = {
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
        }

        file_contents = '\n'.join(rows)

        with self.assertRaises(ValueError):
            await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )
        return

    # TODO mfranklin / vivbak: this test is failing because of change in the parsers
    #   to exclude absolute paths (as absolute paths are NOT in the file map).
    #   I don't know what needs to change to fix this test, except maybe
    #   that the EC parser shouldn't return absolute paths
    # @run_as_sync
    # @patch('metamist.parser.generic_parser.query_async')
    # async def test_missing_fastqs(self, mock_graphql_query):
    #     """
    #     Tests case where the fastq's in the storage do not match the ingested samples.
    #     """
    #     mock_graphql_query.side_effect = self.run_graphql_query_async

    #     rows = [
    #         'HEADER',
    #         '""',
    #         'Application\tExternal ID\tSample Concentration (ng/ul)\tVolume (uL)\tSex\tSample/Name\tReference Genome\tParticipant ID\t',
    #         'App\tEXTID1234\t100\t100\tFemale\t220405_FLUIDX1234\thg38\tPID123',
    #     ]
    #     parser = ExistingCohortParser(
    #         include_participant_column=False,
    #         batch_number='M01',
    #         search_locations=[],
    #         project=self.project_name,
    #         allow_missing_files=False,
    #     )

    #     parser.filename_map = {
    #         'HG3F_2_220405_FLUIDXMISTMATCH1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq': '/path/to/HG3F_2_220405_FLUIDXMISMATCH1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
    #         'HG3F_2_220405_FLUIDXMISMATCH1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq': '/path/to/HG3F_2_220405_FLUIDXMISMATCH1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
    #     }

    #     file_contents = '\n'.join(rows)

    #     with self.assertRaises(ValueError):
    #         await parser.parse_manifest(
    #             StringIO(file_contents), delimiter='\t', dry_run=True
    #         )
    #     return

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.datetime_added')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_existing_row(
        self,
        mock_filesize,
        mock_fileexists,
        mock_datetime_added,
        mock_graphql_query,
    ):
        """
        Tests ingestion for an existing sample.
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        player = ParticipantLayer(self.connection)
        await player.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EXTID1234'},
                    samples=[
                        SampleUpsertInternal(
                            external_ids={PRIMARY_EXTERNAL_ORG: 'EXTID1234'},
                        )
                    ],
                )
            ]
        )

        mock_filesize.return_value = 111
        mock_fileexists.return_value = False
        mock_datetime_added.return_value = datetime.fromisoformat('2022-02-02T22:22:22')

        rows = [
            'HEADER',
            '""',
            'Application\tExternal ID\tSample Concentration (ng/ul)\tVolume (uL)\tSex\tSample/Name\tReference Genome\t',
            'App\tEXTID1234\t100\t100\tFemale\t220405_FLUIDX1234\thg38\t',
        ]
        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=False,
            sequencing_type='genome',
        )

        parser.filename_map = {
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
            'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
        }

        file_contents = '\n'.join(rows)
        summary, _ = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(0, summary.samples.insert)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(1, summary.samples.update)
        self.assertEqual(0, summary.assays.update)

        return

    @run_as_sync
    async def test_get_read_filenames_no_reads_fail(self):
        """Test ValueError is raised when allow_missing_files is False and sequencing groups have no reads"""

        single_row = {Columns.MANIFEST_FLUID_X: ''}

        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=False,
            sequencing_type='genome',
        )
        parser.filename_map = {}

        with self.assertRaises(ValueError):
            # this will raise a ValueError because the allow_missing_files=False,
            # and there are no matching reads in the filename map
            await parser.get_read_filenames(sample_id='', row=single_row)

    @run_as_sync
    async def test_get_read_filenames_no_reads_pass(self):
        """Test when allow_missing_files is True and records with missing fastqs, no ValueError is raised"""

        single_row = {Columns.MANIFEST_FLUID_X: ''}

        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=True,
            sequencing_type='genome',
        )
        parser.filename_map = {}

        with self.assertLogs(level='INFO') as cm:
            read_filenames = await parser.get_read_filenames(
                sample_id='', row=single_row
            )

        self.assertEqual(len(cm.output), 1)
        self.assertIn('No read files found for ', cm.output[0])

        self.assertEqual(len(read_filenames), 0)

    @run_as_sync
    async def test_genome_sequencing_type(self):
        """Test that the sequencing type is set correctly when the --sequencing-type flag is set to 'genome''"""

        # Test with 'genome'
        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=True,
            sequencing_type='genome',
        )
        self.assertEqual(parser.default_sequencing.seq_type, 'genome')

    @run_as_sync
    async def test_exome_sequencing_type(self):
        """Test that the sequencing type is set correctly when the --sequencing-type flag is set to 'exome'"""

        # Test with 'exome'
        parser = ExistingCohortParser(
            include_participant_column=False,
            batch_number='M01',
            search_locations=[],
            project=self.project_name,
            allow_missing_files=True,
            sequencing_type='exome',
        )
        self.assertEqual(parser.default_sequencing.seq_type, 'exome')

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    @patch('metamist.parser.cloudhelper.CloudHelper.datetime_added')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_exists')
    @patch('metamist.parser.cloudhelper.CloudHelper.file_size')
    async def test_sequencing_type_in_assay_meta(
        self,
        mock_filesize,
        mock_fileexists,
        mock_datetime_added,
        mock_graphql_query,
    ):
        """Test that the sequencing type is set correctly when the --sequencing-type flag is set to 'genome' or 'exome'"""

        mock_graphql_query.side_effect = self.run_graphql_query_async

        mock_filesize.return_value = 111
        mock_fileexists.return_value = False
        mock_datetime_added.return_value = datetime.fromisoformat('2022-02-02T22:22:22')

        rows = [
            'HEADER',
            '""',
            'Application\tExternal ID\tSample Concentration (ng/ul)\tVolume (uL)\tSex\tSample/Name\tReference Genome\tParticipant ID\t',
            'App\tEXTID1234\t100\t100\tFemale\t220405_FLUIDX1234\thg38\tPID123',
        ]

        for sequencing_type in ['genome', 'exome']:
            with self.subTest(sequencing_type=sequencing_type):
                parser = ExistingCohortParser(
                    include_participant_column=False,
                    batch_number='M01',
                    search_locations=[],
                    project=self.project_name,
                    allow_missing_files=False,
                    sequencing_type=sequencing_type,
                )
                parser.filename_map = {
                    'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
                    'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
                }

                file_contents = '\n'.join(rows)
                participants: list[ParsedParticipant]
                _, participants = await parser.parse_manifest(
                    StringIO(file_contents), delimiter='\t', dry_run=True
                )

                sample_to_add = participants[0].samples[0]
                expected_sequence_dict = {
                    'reference_genome': 'hg38',
                    'platform': 'App',
                    'concentration': 100,
                    'volume': 100,
                    'fluid_x_tube_id': '220405_FLUIDX1234',
                    'reads_type': 'fastq',
                    'reads': [
                        {
                            'location': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
                            'basename': 'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R1.fastq',
                            'class': 'File',
                            'checksum': None,
                            'size': 111,
                            'datetime_added': '2022-02-02T22:22:22',
                        },
                        {
                            'location': '/path/to/HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
                            'basename': 'HG3F_2_220405_FLUIDX1234_Homo-sapiens_AAC-TAT_R_220208_VB_BLAH_M002_R2.fastq',
                            'class': 'File',
                            'checksum': None,
                            'size': 111,
                            'datetime_added': '2022-02-02T22:22:22',
                        },
                    ],
                    'sequencing_platform': 'illumina',
                    'sequencing_technology': 'short-read',
                    'sequencing_type': f'{sequencing_type}',
                    'batch': 'M01',
                }
                assay = sample_to_add.sequencing_groups[0].assays[0]
                self.maxDiff = None
                self.assertDictEqual(expected_sequence_dict, assay.meta)
        return



import unittest
from datetime import datetime
from enum import Enum

import google.cloud.bigquery as bq
import pytz

from db.python.tables.bq.function_bq_filter import FunctionBQFilter
from models.models import BillingColumn


class BGFunFilterTestEnum(str, Enum):
    """Simple Enum classs"""

    ID = 'id'
    VALUE = 'value'


class TestFunctionBQFilter(unittest.TestCase):
    """Test function filters SQL generation"""

    def test_empty_output(self):
        """Test that function filter converts to SQL as expected"""
        filter_ = FunctionBQFilter(
            name='test_sql_func',
            implementation='CREATE TEMP FUNCTION test_sql_func() AS (SELECT 1);',
        )

        # no params are present, should return empty SQL and Param list
        filter_.to_sql(BillingColumn.LABELS)

        self.assertEqual([], filter_.func_sql_parameters)
        self.assertEqual('', filter_.func_where)

    def test_one_param_output(self):
        """Test that function filter converts to SQL as expected"""
        filter_ = FunctionBQFilter(
            name='test_sql_func',
            implementation='CREATE TEMP FUNCTION test_sql_func() AS (SELECT 1);',
        )

        # no params are present, should return empty SQL and Param list
        filter_.to_sql(BillingColumn.LABELS, {'label_name': 'Some Value'})

        self.assertEqual(
            [
                bq.ScalarQueryParameter('param1', 'STRING', 'label_name'),
                bq.ScalarQueryParameter('value1', 'STRING', 'Some Value'),
            ],
            filter_.func_sql_parameters,
        )
        self.assertEqual(
            '(test_sql_func(labels,@param1) = @value1)', filter_.func_where
        )

    def test_two_param_default_operator_output(self):
        """Test that function filter converts to SQL as expected"""
        filter_ = FunctionBQFilter(
            name='test_sql_func',
            implementation='CREATE TEMP FUNCTION test_sql_func() AS (SELECT 1);',
        )

        # no params are present, should return empty SQL and Param list
        filter_.to_sql(BillingColumn.LABELS, {'label_name1': 10, 'label_name2': 123.45})

        self.assertEqual(
            [
                bq.ScalarQueryParameter('param1', 'STRING', 'label_name1'),
                bq.ScalarQueryParameter('value1', 'INT64', 10),
                bq.ScalarQueryParameter('param2', 'STRING', 'label_name2'),
                bq.ScalarQueryParameter('value2', 'FLOAT64', 123.45),
            ],
            filter_.func_sql_parameters,
        )
        self.assertEqual(
            (
                '(test_sql_func(labels,@param1) = @value1 '
                'AND test_sql_func(labels,@param2) = @value2)'
            ),
            filter_.func_where,
        )

    def test_two_param_operator_or_output(self):
        """Test that function filter converts to SQL as expected"""
        filter_ = FunctionBQFilter(
            name='test_sql_func',
            implementation='CREATE TEMP FUNCTION test_sql_func() AS (SELECT 1);',
        )

        # no params are present, should return empty SQL and Param list
        filter_.to_sql(
            BillingColumn.LABELS,
            {
                'label_name1': BGFunFilterTestEnum.ID,
                'label_name2': datetime(2024, 1, 1),
            },
            'OR',
        )

        self.assertEqual(
            [
                bq.ScalarQueryParameter('param1', 'STRING', 'label_name1'),
                bq.ScalarQueryParameter('value1', 'STRING', 'id'),
                bq.ScalarQueryParameter('param2', 'STRING', 'label_name2'),
                bq.ScalarQueryParameter('value2', 'STRING', '2024-01-01T00:00:00'),
            ],
            filter_.func_sql_parameters,
        )
        self.assertEqual(
            (
                '(test_sql_func(labels,@param1) = @value1 '
                'OR test_sql_func(labels,@param2) = TIMESTAMP(@value2))'
            ),
            filter_.func_where,
        )

    def test_two_param_datime_with_zone_output(self):
        """Test that function filter converts to SQL as expected"""
        filter_ = FunctionBQFilter(
            name='test_sql_func',
            implementation='CREATE TEMP FUNCTION test_sql_func() AS (SELECT 1);',
        )

        # no params are present, should return empty SQL and Param list
        NYC = pytz.timezone('America/New_York')
        SYD = pytz.timezone('Australia/Sydney')
        filter_.to_sql(
            BillingColumn.LABELS,
            {
                'label_name1': datetime(2024, 1, 1, tzinfo=pytz.UTC).astimezone(NYC),
                'label_name2': datetime(2024, 1, 1, tzinfo=pytz.UTC).astimezone(SYD),
            },
            'AND',
        )

        self.assertEqual(
            [
                bq.ScalarQueryParameter('param1', 'STRING', 'label_name1'),
                bq.ScalarQueryParameter(
                    'value1', 'STRING', '2023-12-31T19:00:00-05:00'
                ),
                bq.ScalarQueryParameter('param2', 'STRING', 'label_name2'),
                bq.ScalarQueryParameter(
                    'value2', 'STRING', '2024-01-01T11:00:00+11:00'
                ),
            ],
            filter_.func_sql_parameters,
        )
        self.assertEqual(
            (
                '(test_sql_func(labels,@param1) = TIMESTAMP(@value1) '
                'AND test_sql_func(labels,@param2) = TIMESTAMP(@value2))'
            ),
            filter_.func_where,
        )



from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.family import FamilyLayer
from db.python.layers.participant import ParticipantLayer
from models.models import PRIMARY_EXTERNAL_ORG, ParticipantUpsertInternal


class TestPedigree(DbIsolatedTest):
    """Pedigree testing methods"""

    @run_as_sync
    async def test_import_get_pedigree(self):
        """Test import + get pedigree"""
        fl = FamilyLayer(self.connection)

        rows: list[list[str]] = [
            ['FAM01', 'EX01_father', '', '', '1', '1'],
            ['FAM01', 'EX01_mother', '', '', '2', '1'],
            ['FAM01', 'EX01_subject', 'EX01_father', 'EX01_mother', '1', '2'],
        ]

        await fl.import_pedigree(
            header=None, rows=rows, create_missing_participants=True
        )

        pedigree_dicts = await fl.get_pedigree(
            project=self.connection.project_id,
            replace_with_participant_external_ids=True,
            replace_with_family_external_ids=True,
        )

        by_key = {r['individual_id']: r for r in pedigree_dicts}

        self.assertEqual(3, len(pedigree_dicts))
        father = by_key['EX01_father']
        mother = by_key['EX01_mother']
        subject = by_key['EX01_subject']

        self.assertIsNone(father['paternal_id'])
        self.assertIsNone(mother['paternal_id'])
        self.assertEqual('EX01_father', subject['paternal_id'])
        self.assertEqual('EX01_mother', subject['maternal_id'])

    @run_as_sync
    async def test_pedigree_without_family(self):
        """
        Test getting pedigree where participants do not belong to a family
        """
        pl = ParticipantLayer(self.connection)
        fl = FamilyLayer(self.connection)

        await pl.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'},
                reported_sex=1,
            )
        )
        await pl.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX02'}, reported_sex=None
            )
        )

        rows = await fl.get_pedigree(
            project=self.connection.project_id,
            include_participants_not_in_families=True,
            replace_with_participant_external_ids=True,
        )

        by_id = {r['individual_id']: r for r in rows}
        self.assertEqual(2, len(rows))
        self.assertEqual(1, by_id['EX01']['sex'])
        self.assertIsNone(by_id['EX02']['sex'])



import uuid
from test.testbase import DbIsolatedTest, run_as_sync

from db.python.tables.project import (
    GROUP_NAME_MEMBERS_ADMIN,
    GROUP_NAME_PROJECT_CREATORS,
    ProjectPermissionsTable,
)
from db.python.utils import Forbidden
from models.models.project import (
    FullWriteAccessRoles,
    ProjectMemberRole,
    ProjectMemberUpdate,
    ReadAccessRoles,
)


class TestGroupAccess(DbIsolatedTest):
    """
    Test project access permissions + newish internal group implementation
    """

    @run_as_sync
    async def setUp(self):
        """Setup tests"""
        super().setUp()

        # specifically required to test permissions
        self.pttable = ProjectPermissionsTable(self.connection)

    async def _add_group_member_direct(self, group_name: str):
        """
        Helper method to directly add members to group with name
        """
        members_admin_group = await self.connection.connection.fetch_val(
            'SELECT id FROM `group` WHERE name = :name',
            {'name': group_name},
        )
        await self.connection.connection.execute(
            """
            INSERT INTO group_member (group_id, member, audit_log_id)
            VALUES (:group_id, :member, :audit_log_id);
            """,
            {
                'group_id': members_admin_group,
                'member': self.author,
                'audit_log_id': await self.audit_log_id(),
            },
        )

    @run_as_sync
    async def test_project_creators_failed(self):
        """
        Test that a user without permission cannot create a project
        """
        with self.assertRaises(Forbidden):
            await self.pttable.create_project(
                'another-test-project', 'another-test-project', self.author
            )

    @run_as_sync
    async def test_project_create_succeed(self):
        """
        Test that a user with permission can create a project,
        and that read/write groups are created
        """
        await self._add_group_member_direct(GROUP_NAME_PROJECT_CREATORS)
        g = str(uuid.uuid4())

        project_id = await self.pttable.create_project(g, g, self.author)

        # pylint: disable=protected-access
        project_id_map, _ = await self.pttable.get_projects_accessible_by_user(
            user=self.author
        )

        project = project_id_map.get(project_id)
        assert project
        self.assertEqual(project.name, g)


class TestProjectAccess(DbIsolatedTest):
    """Test project access methods directly"""

    @run_as_sync
    async def setUp(self):
        """Setup tests"""
        super().setUp()

        # specifically required to test permissions
        self.pttable = ProjectPermissionsTable(self.connection)

    async def _add_group_member_direct(
        self,
        group_name: str,
    ):
        """
        Helper method to directly add members to group with name
        """
        members_admin_group = await self.connection.connection.fetch_val(
            'SELECT id FROM `group` WHERE name = :name',
            {'name': group_name},
        )
        await self.connection.connection.execute(
            """
            INSERT INTO group_member (group_id, member, audit_log_id)
            VALUES (:group_id, :member, :audit_log_id);
            """,
            {
                'group_id': members_admin_group,
                'member': self.author,
                'audit_log_id': await self.audit_log_id(),
            },
        )

    @run_as_sync
    async def test_no_project_access(self):
        """
        Test that a user without permission cannot access a project
        """
        await self._add_group_member_direct(GROUP_NAME_PROJECT_CREATORS)
        g = str(uuid.uuid4())

        project_id = await self.pttable.create_project(g, g, self.author)
        with self.assertRaises(Forbidden):
            self.connection.check_access_to_projects_for_ids(
                project_ids=[project_id], allowed_roles=ReadAccessRoles
            )

        with self.assertRaises(Forbidden):
            self.connection.get_and_check_access_to_projects_for_names(
                project_names=[g], allowed_roles=ReadAccessRoles
            )

    @run_as_sync
    async def test_project_access_success(self):
        """
        Test that a user with permission CAN access a project
        """
        await self._add_group_member_direct(GROUP_NAME_PROJECT_CREATORS)
        await self._add_group_member_direct(GROUP_NAME_MEMBERS_ADMIN)

        g = str(uuid.uuid4())

        pid = await self.pttable.create_project(g, g, self.author)

        project_id_map, _ = await self.pttable.get_projects_accessible_by_user(
            user=self.author
        )
        project = project_id_map.get(pid)
        assert project
        await self.pttable.set_project_members(
            project=project,
            members=[ProjectMemberUpdate(member=self.author, roles=['reader'])],
        )

        project_for_id = self.connection.get_and_check_access_to_projects_for_ids(
            project_ids=[pid], allowed_roles=ReadAccessRoles
        )
        user_project_for_id = next(p for p in project_for_id)
        self.assertEqual(pid, user_project_for_id.id)

        project_for_name = self.connection.get_and_check_access_to_projects_for_names(
            project_names=[g], allowed_roles=ReadAccessRoles
        )
        user_project_for_name = next(p for p in project_for_name)
        self.assertEqual(g, user_project_for_name.name)

    @run_as_sync
    async def test_project_access_insufficient(self):
        """
        Test that a user with access to a project will be disallowed if their access is
        not sufficient
        """
        await self._add_group_member_direct(GROUP_NAME_PROJECT_CREATORS)
        await self._add_group_member_direct(GROUP_NAME_MEMBERS_ADMIN)

        g = str(uuid.uuid4())

        pid = await self.pttable.create_project(g, g, self.author)

        project_id_map, _ = await self.pttable.get_projects_accessible_by_user(
            user=self.author
        )
        project = project_id_map.get(pid)
        assert project
        # Give the user read access to the project
        await self.pttable.set_project_members(
            project=project,
            members=[ProjectMemberUpdate(member=self.author, roles=['reader'])],
        )

        # But require Write access

        with self.assertRaises(Forbidden):
            self.connection.check_access_to_projects_for_ids(
                project_ids=[project.id], allowed_roles=FullWriteAccessRoles
            )

        with self.assertRaises(Forbidden):
            self.connection.get_and_check_access_to_projects_for_names(
                project_names=[g], allowed_roles=FullWriteAccessRoles
            )

    @run_as_sync
    async def test_get_my_projects(self):
        """
        Test that a user with permission only has MY projects
        """
        await self._add_group_member_direct(GROUP_NAME_PROJECT_CREATORS)
        await self._add_group_member_direct(GROUP_NAME_MEMBERS_ADMIN)

        g = str(uuid.uuid4())

        pid = await self.pttable.create_project(g, g, self.author)

        project_id_map, _ = await self.pttable.get_projects_accessible_by_user(
            user=self.author
        )
        project = project_id_map.get(pid)
        assert project
        # Give the user read access to the project
        await self.pttable.set_project_members(
            project=project,
            members=[ProjectMemberUpdate(member=self.author, roles=['contributor'])],
        )

        (
            project_id_map,
            project_name_map,
        ) = await self.pttable.get_projects_accessible_by_user(user=self.author)

        # Get projects with at least a read access role
        my_projects = self.connection.projects_with_role(
            {ProjectMemberRole.contributor}
        )
        print(my_projects)

        self.assertEqual(len(project_id_map), len(project_name_map))
        self.assertEqual(len(my_projects), 1)
        self.assertEqual(pid, my_projects[0].id)
